#
"""test_week - Test ISO-8601 week handling"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date
from typing import Tuple

from pytest import mark

from rdial.utils import iso_week_to_date


@mark.parametrize('year, week, expected', [
    (2007, 1, (date(2007, 1, 1), date(2007, 1, 8))),  # Year starts same day
    (2009, 53, (date(2009, 12, 28), date(2010, 1, 4))),  # ISO year spans 2010
    (2013, 52, (date(2013, 12, 23), date(2013, 12, 30))),
])
def test_iso_week_to_date(year: int, week: int, expected: Tuple[date, date]):
    assert iso_week_to_date(year, week) == expected



#
"""events - Event models for rdial."""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#                        Nathan McGregor <nathan.mcgregor@astrium.eads.net>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import csv
import datetime
import glob
import inspect
import operator
import os
import pickle
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Union

import click

from jnrbase import iso_8601, xdg_basedir

from . import utils


class RdialDialect(csv.unix_dialect):  # pylint: disable=too-few-public-methods

    """CSV dialect for rdial data files."""

    quoting = csv.QUOTE_MINIMAL
    strict = True


class TaskNotRunningError(utils.RdialError):

    """Exception for calling mutators when a task is not running."""


class TaskRunningError(utils.RdialError):

    """Exception for starting task when a task is already running."""


class TaskNotExistError(utils.RdialError):

    """Exception for attempting to operate on a non-existing task."""


class Event:

    """Base object for handling database event."""

    def __init__(self, __task: str,
                 start: Optional[Union[datetime.datetime, str]] = None,
                 delta: Optional[Union[datetime.timedelta, str]] = None,
                 message: Optional[str] = '') -> None:
        """Initialise a new ``Event`` object.

        Args:
            __task: Task name to tracking
            start: Start time for event
            delta: Duration for event
            message: Message to attach to event

        """
        self.task = __task
        if isinstance(start, datetime.datetime):
            if start.tzinfo:
                raise ValueError(f'Must be a naive datetime {start!r}')
            self.start = start
        else:
            self.start = iso_8601.parse_datetime(start).replace(tzinfo=None)
        if isinstance(delta, datetime.timedelta):
            self.delta = delta
        else:
            self.delta = iso_8601.parse_delta(delta)
        self.message = message

    def __eq__(self, __other: 'Event') -> bool:
        """Comare ``Event`` objects for equality.

        Args:
            __other: Object to test equality against

        Returns:
            True if objects are equal
        """
        return self.task == __other.task and self.start == __other.start \
            and self.delta == __other.delta and self.message == __other.message

    def __ne__(self, __other: 'Event') -> bool:
        """Comare ``Event`` objects for inequality.

        Args:
            __other: Object to test inequality against

        Returns:
            True if objects are not equal
        """
        return not self == __other

    def __repr__(self) -> str:
        """Self-documenting string representation.

        Returns:
            Event representation suitable for :func:`eval`
        """
        return 'Event({!r}, {!r}, {!r}, {!r})'.format(
            self.task, iso_8601.format_datetime(self.start) + 'Z',
            iso_8601.format_delta(self.delta), self.message)

    def writer(self) -> Dict[str, Optional[str]]:
        """Prepare object for export.

        Returns:
            Event data for object storage

        """
        return {
            'start': iso_8601.format_datetime(self.start) + 'Z',
            'delta': iso_8601.format_delta(self.delta),
            'message': self.message,
        }

    def running(self) -> Union[str, bool]:
        """Check if event is running.

        Returns:
            Event name, if running

        """
        if self.delta == datetime.timedelta(0):
            return self.task
        return False

    def stop(self, message: Optional[str] = None, force: bool = False) -> None:
        """Stop running event.

        Args:
            message: Message to attach to event
            force: Re-stop a previously stopped event

        Raises:
            TaskNotRunningError: Event not running

        """
        if not force and self.delta:
            raise TaskNotRunningError('No task running!')
        self.delta = datetime.datetime.utcnow() - self.start
        self.message = message
FIELDS = list(inspect.signature(Event).parameters.keys())[1:]  # NOQA


class Events(list):  # pylint: disable=too-many-public-methods

    """Container for database events."""

    def __init__(self, __iterable: Optional[List[Event]] = None,
                 backup: bool = True) -> None:
        """Initialise a new ``Events`` object.

        Args:
            __iterable: Objects to add to container
            backup: Whether to create backup files

        """
        super(Events, self).__init__(__iterable if __iterable else [])
        self.backup = backup
        self._dirty = set()

    def __repr__(self) -> str:
        """Self-documenting string representation.

        Returns:
            Events representation suitable for :func:`eval`
        """
        return 'Events({})'.format(super(self.__class__, self).__repr__())

    @property
    def dirty(self) -> Iterable[str]:
        """Modified tasks requiring sync against storage."""
        return self._dirty

    @dirty.setter
    def dirty(self, __value: str):
        """Mark task as needing sync.

        Args:
            __value: Task to mark as dirty

        """
        self._dirty.add(__value)

    @dirty.deleter
    def dirty(self):
        """Mark dirty queue as flushed."""
        self._dirty = set()

    @staticmethod
    def read(__directory: str, backup: bool = True,
             write_cache: bool = True) -> 'Events':
        """Read and parse database.

        .. note::

            Assumes a new :obj:`Events` object should be created if the
            directory is missing.

        Args:
            __directory: Location to read database files from
            backup: Whether to create backup files
            write_cache: Whether to write cache files

        Returns:
            Parsed events database

        """
        if not os.path.exists(__directory):
            return Events(backup=backup)
        events = []
        xdg_cache_dir = xdg_basedir.user_cache('rdial')
        cache_dir = os.path.join(xdg_cache_dir, __directory.replace('/', '_'))
        if write_cache and not os.path.isdir(cache_dir):
            os.makedirs(cache_dir)
            with click.open_file(f'{xdg_cache_dir}/CACHEDIR.TAG', 'w') as f:
                f.writelines([
                    'Signature: 8a477f597d28d172789f06886806bc55\n',
                    '# This file is a cache directory tag created by rdial.\n',
                    '# For information about cache directory tags, see:\n',
                    '#   http://www.brynosaurus.com/cachedir/\n',
                ])
        for fname in glob.glob(f'{__directory}/*.csv'):
            task = os.path.basename(fname)[:-4]
            cache_file = os.path.join(cache_dir, task) + '.pkl'
            evs = None
            if os.path.exists(cache_file) and utils.newer(cache_file, fname):
                try:
                    # UnicodeDecodeError must be caught for the Python 2 to
                    # 3 upgrade path.
                    with click.open_file(cache_file, 'rb') as f:
                        cache = pickle.load(f)
                except (pickle.UnpicklingError, EOFError, ImportError,
                        UnicodeDecodeError):
                    pass
                else:
                    try:
                        assert cache['version'] == 1
                    except TypeError:
                        os.unlink(cache_file)
                    else:
                        evs = cache['events']
            if evs is None:
                with click.open_file(fname, encoding='utf-8') as f:
                    # We're not using the prettier DictReader here as it is
                    # *significantly* slower for large data files (~5x).
                    reader = csv.reader(f, dialect=RdialDialect)
                    assert next(reader) == FIELDS, \
                        'Invalid data {!r}'.format(
                            click.format_filename(fname))
                    evs = [Event(task, *row)  # pylint: disable=star-args
                           for row in reader]
                if write_cache:
                    with click.open_file(cache_file, 'wb', atomic=True) as f:
                        pickle.dump({'version': 1, 'events': evs}, f,
                                    pickle.HIGHEST_PROTOCOL)
            events.extend(evs)
        return Events(sorted(events, key=operator.attrgetter('start')))

    def write(self, __directory: str) -> None:
        """Write database file.

        Args:
            __directory: Location to write database files to

        """
        if not self.dirty:
            return
        if not os.path.isdir(__directory):
            os.makedirs(__directory)

        for task in self.dirty:
            task_file = f'{__directory}/{task}.csv'
            events = self.for_task(task)
            with click.utils.LazyFile(task_file, 'w', atomic=True) as temp:
                writer = csv.DictWriter(temp, FIELDS, dialect=RdialDialect)
                writer.writeheader()
                for event in events:
                    writer.writerow(event.writer())
                if self.backup and os.path.exists(task_file):
                    os.rename(task_file, f'{task_file}~')
        del self.dirty

    def tasks(self) -> List[str]:
        """Generate a list of tasks in the database.

        Returns:
            Names of tasks in database

        """
        return sorted({event.task for event in self})

    def last(self) -> Optional[Event]:
        """Return current/last event.

        This handles the empty database case by returning ``None``

        Returns:
            Last recorded event

        """
        if len(self) > 0:
            return self[-1]
        else:
            return None

    def running(self) -> Union[str, bool]:
        """Check if an event is running.

        We return the running task, if one exists, for easy access.

        Returns:
            Running event, if an event running

        """
        last = self.last()
        return last.running() if last else False

    def start(self, __task: str, new: bool = False,
              start: Union[datetime.datetime, str] = '') -> None:
        """Start a new event.

        Args:
            __task: Task name to tracking
            new: Whether to create a new task
            start: ISO-8601 start time for event

        Raises:
            TaskRunningError: An event is already running

        """
        if not new and __task not in self.tasks():
            raise TaskNotExistError(
                f'Task {__task} does not exist!  Use “--new” to create it')
        running = self.running()
        if running:
            raise TaskRunningError(f'Running task {running}!')
        last = self.last()
        if last and start and last.start + last.delta > start:
            raise TaskRunningError('Start date overlaps previous task!')
        self.append(Event(__task, start))
        self.dirty = __task

    def stop(self, message: Optional[str] = None, force: bool = False) -> None:
        """Stop running event.

        Args:
            message: Message to attach to event
            force: Re-stop a previously stopped event

        Raises:
            TaskNotRunningError: No task running!

        """
        if not force and not self.running():
            raise TaskNotRunningError('No task running!')
        self.last().stop(message, force)
        self.dirty = self.last().task

    def filter(self, __filt: Callable[[Event, ], bool]) -> 'Events':
        """Apply filter to events.

        Args:
            __filt: Function to filter with

        Returns:
            Events matching given filter function

        """
        return Events(x for x in self if __filt(x))

    def for_task(self, __task: str) -> 'Events':
        """Filter events for a specific task.

        Args:
            __task: Task name to filter on

        Returns:
            Events marked with given task name

        """
        return self.filter(lambda x: x.task == __task)

    def for_date(self, year: int, month: Optional[int] = None,
                 day: Optional[int] = None) -> 'Events':
        """Filter events for a specific date.

        Args:
            year: Year to filter on
            month: Month to filter on
            day: Day to filter on

        Returns:
            Events occurring within specified date

        """
        events = self.filter(lambda x: x.start.year == year)
        if month:
            events = events.filter(lambda x: x.start.month == month)
        if day:
            events = events.filter(lambda x: x.start.day == day)
        return events

    def for_week(self, __year: int, __week: int) -> 'Events':
        """Filter events for a specific ISO-8601 week.

        Args:
            __year: Year to filter events on
            __week: ISO-8601 month number to filter events on

        Returns:
            Events occurring in given ISO-8601 week
        """
        start, end = utils.iso_week_to_date(__year, __week)
        return self.filter(lambda x: start <= x.start.date() < end)

    def sum(self) -> datetime.timedelta:
        """Sum duration of all events.

        Returns:
            Sum of all event deltas

        """
        return sum((x.delta for x in self), datetime.timedelta(0))

    @staticmethod
    @contextlib.contextmanager
    def wrapping(__directory: str, backup: bool = True,
                 write_cache: bool = True) -> Iterator['Events']:
        """Convenience context handler to manage reading and writing database.

        Args:
            __directory: Database location
            backup: Whether to create backup files
            write_cache: Whether to write cache files

        """
        events = Events.read(__directory, backup, write_cache)
        yield events
        if events.dirty:
            events.write(__directory)



#
"""rdial - Simple time tracking for simple people.

A simple time tracking tool, with no frills and no fizzy coating.
"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from . import _version


__version__ = _version.dotted
__date__ = _version.date
__copyright__ = 'Copyright © 2011-2017  James Rowe <jnrowe@gmail.com>'



#
"""test_utils - Test utility functions"""
# Copyright © 2014-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from time import sleep
from typing import Optional

from jnrbase.attrdict import AttrDict
from pytest import mark

from rdial.utils import (newer, read_config, remove_current, term_link,
                         write_current)


def test_read_config_local():
    conf = read_config('tests/data/local.ini')
    assert conf['local test'].getboolean('read')


def test_handle_current(tmpdir):
    globs = AttrDict(directory=tmpdir.strpath)
    bare = lambda globs, task: True
    write_current(bare)(globs, task='test')
    assert tmpdir.join('.current').exists()
    remove_current(bare)(globs, task='test')
    assert not tmpdir.join('.current').exists()
    # check idempotent...
    remove_current(bare)(globs, task='test')


def test_newer(tmpdir):
    f1 = tmpdir.join('file1').ensure()
    sleep(0.1)
    f2 = tmpdir.join('file2').ensure()
    assert newer(f2.strpath, f1.strpath)
    assert not newer(f1.strpath, f2.strpath)
    assert not newer(f1.strpath, f1.strpath)


@mark.parametrize('target, name, result', [
    ('pypi://rdial', 'this package',
     '\x1b]8;;pypi://rdial\x07this package\x1b]8;;\x07'),
    ('pypi://rdial', None, '\x1b]8;;pypi://rdial\x07rdial\x1b]8;;\x07'),
])
def test_term_link(target: str, name: Optional[str], result: str):
    assert term_link(target, name) == result



#
"""test_datetime - Test ISO-8601 handling"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timedelta

from pytest import mark

from rdial.utils import parse_datetime_user


@mark.parametrize('string, delta', [
    ('5 minutes ago', timedelta(minutes=5)),
    ('1 hour ago -5 minutes', timedelta(hours=1, minutes=5)),
])
def test_parse_datetime_via_date_command(string: str, delta: timedelta):
    now = datetime.utcnow().replace(microsecond=0)
    # Accept a 2.5 second smudge window
    assert parse_datetime_user(string) >= now - delta
    assert parse_datetime_user(string) < now - delta + timedelta(seconds=2.5)



#
"""test_repr - Test __repr__ methods for self-reproducibility"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from rdial.events import Event, Events


def eval_repr(obj: object):
    """Regenerate an object from its ``__repr__`` output.

    Args:
        obj: Object to evaluate

    """
    return repr(eval(repr(obj)))


class ReprTest(unittest.TestCase):

    """``repr()`` should return self-documenting string."""

    def test_event(self):
        ev = Event('task', '2011-05-05T11:23:48Z', 'PT01H00M00S')
        assert repr(ev) == eval_repr(ev)

    def test_event_no_delta(self):
        ev = Event('task', '2011-05-05T11:23:48Z', '')
        assert repr(ev) == eval_repr(ev)

    def test_event_with_message(self):
        ev = Event('task', '2011-05-05T11:23:48Z', 'PT01H00M00S', 'message')
        assert repr(ev) == eval_repr(ev)

    def test_events(self):
        ev1 = Event('task', '2011-05-05T11:23:48Z', 'PT01H00M00S')
        ev2 = Event('task', '2011-05-05T12:23:48Z', 'PT00H30M00S')
        events = Events([ev1, ev2])
        assert repr(events) == eval_repr(events)



#
"""test_setup_py - Test setup.py functionality"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

from pytest import mark

import setup


@mark.parametrize('input, expected', [
    ('plain', ['tabulate', ]),
    ('recurse', ['click', 'tabulate', ]),
])
def test_parse_requires(input: str, expected: List[str]):
    requires = setup.parse_requires(f'../tests/data/requires/{input}.txt')
    assert requires == expected



#
"""test_event_filter - Test event filter handling"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List, Optional

from jnrbase.attrdict import AttrDict
from pytest import mark

from rdial.cmdline import filter_events
from rdial.events import Events


def test_fetch_events_for_task():
    events = Events.read('tests/data/test', write_cache=False)
    assert len(events.for_task('task2')) == 1


@mark.parametrize('date, expected', [
    ({'year': 2011, }, 2),
    ({'year': 2011, 'month': 1}, 1),
    ({'year': 2011, 'month': 3, 'day': 1}, 1),
    ({'year': 2011, 'month': 3, 'day': 31}, 0),
])
def test_fetch_events_for_date(date: Dict[str, int], expected: int):
    events = Events.read('tests/data/date_filtering', write_cache=False)
    assert len(events.for_date(**date)) == expected


def test_fetch_events_for_week():
    events = Events.read('tests/data/date_filtering', write_cache=False)
    assert len(events.for_week(2011, 9)) == 1


@mark.parametrize('task, result', [
    (None, ['task', 'task2']),
    ('task', ['task', ]),
])
def test_filter_events_by_task(task: Optional[str], result: List[str]):
    globs = AttrDict(directory='tests/data/test', cache=False)
    evs = filter_events(globs, task)
    assert evs.tasks() == result



#
"""test_cmdline - Test command line functionality"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from shutil import copytree
from typing import Callable, Optional

from click import (BadParameter, Context, Option, command, echo, group,
                   option, pass_context, pass_obj)
from click.testing import CliRunner
from pytest import mark, raises

from rdial.cmdline import (HiddenGroup, StartTimeParamType, TaskNameParamType,
                           cli, get_stop_message, hidden, main, task_option)
from rdial.events import Event, TaskNotRunningError, TaskRunningError


@mark.parametrize('string, expected', [
    ('valid_name', True),
    ('also-valid-name', True),
    ('.invalid_name', BadParameter),
    ('valid.name', True),
    ('invalid/name', BadParameter),
    ('', BadParameter),
    ('x' * 256, BadParameter),
    ('-bad-start', BadParameter),
])
def test_task_name_validity(string: str, expected: Optional[BadParameter]):
    p = TaskNameParamType()
    if expected is True:
        assert p.convert(string, None, None) == string
    else:
        with raises(expected):
            p.convert(string, None, None)


def test_task_name_from_dir(tmpdir):
    @task_option
    @command()
    def cli(task):
        echo(task)

    runner = CliRunner()
    result = runner.invoke(cli, ['--from-dir', ])
    assert result.output == 'rdial\n'
    with tmpdir.mkdir('new_dir').as_cwd():
        result = runner.invoke(cli, ['--from-dir', ])
        assert result.output.endswith('new_dir\n')


@mark.parametrize('string, expected', [
    ('yesterday', True),
    ('', True),
    ('2011-05-04T09:15:00Z', True),
    ('AB1 time', BadParameter),
])
def test_start_time_validity(string: str, expected: Optional[BadParameter]):
    p = StartTimeParamType()
    if expected is True:
        assert isinstance(p.convert(string, None, None), datetime)
    else:
        with raises(expected):
            p.convert(string, None, None)


def test_HiddenGroup():
    @group(cls=HiddenGroup)
    def cli():
        pass

    @cli.command()
    def now_you_see_me():
        pass

    @hidden
    @cli.command()
    def now_you_dont():
        pass

    runner = CliRunner()
    result = runner.invoke(cli, ['--help'])
    lines = [s.strip() for s in result.output.splitlines()]
    assert 'now_you_see_me' in lines
    assert 'now_you_dont' not in lines


@mark.parametrize('edit_func, message', [
    (lambda s, **kwargs: s,
     'finished'),
    (lambda _, **kwargs: None,
     ''),
])
def test_get_stop_message(edit_func: Callable, message: str, monkeypatch):
    monkeypatch.setattr('click.edit', edit_func)
    ev = Event('task', '2011-05-04T09:30:00Z', '', message)
    assert get_stop_message(ev) == message


def test_get_stop_message_template(monkeypatch):
    output = []
    monkeypatch.setattr('click.edit',
                        lambda s, **kwargs: output.extend(s.splitlines()))
    ev = Event('task', '2011-05-04T09:30:00Z')
    get_stop_message(ev)
    assert '# Task “task” started 2011-05-04T09:30:00Z' in output


@mark.parametrize('config, result', [
    ('color', True),
    ('no_color', False),
])
def test_colour_for_u_deficient(config: str, result: bool):
    @cli.command()
    @pass_obj
    def raise_config(obj):
        raise ValueError(obj)

    runner = CliRunner()
    with raises(ValueError) as excinfo:
        runner.invoke(cli,
                      ['--config', f'tests/data/{config}.ini', 'raise_config'],
                      catch_exceptions=False)
    assert excinfo.value.args[0].colour is result


def test_command_defaults():
    @cli.command()
    @option('--choice')
    @pass_context
    def raise_context(ctx: Context, choice: Option):
        raise ValueError(ctx)

    runner = CliRunner()
    with raises(ValueError) as excinfo:
        runner.invoke(cli,
                      ['--config', 'tests/data/defaults.ini',
                       'raise_context', ],
                      catch_exceptions=False)
    defaults = excinfo.value.args[0].default_map
    assert defaults['choice'] == 'questionable'


def test_bug_data():
    runner = CliRunner()
    result = runner.invoke(cli, ['bug-data', ])
    assert result.exit_code == 0
    assert '{' not in result.output
    assert '}' not in result.output


def test_start_event(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'start', 'task'])
    assert result.exit_code == 0
    assert result.output == ''


def test_restart_event(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'start',
                                 '--continue'])
    assert result.exit_code == 0
    assert result.output == ''


def test_stop_event(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'stop'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output


def test_stop_event_with_file_message(tmpdir):
    test_dir = tmpdir.join('test').strpath
    msg_file = tmpdir.join('message').strpath
    with open(msg_file, 'w') as f:
        f.write('stopping message')
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'stop', '-F',
                                 msg_file])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert f.read().endswith('stopping message\n')


def test_stop_event_not_running(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'stop'])
    assert isinstance(result.exception, TaskNotRunningError)
    assert result.exception.args[0] == 'No task running!'


def test_stop_event_running_amend(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'stop', '--amend'])
    assert isinstance(result.exception, TaskRunningError)
    assert result.exception.args[0] == 'Can’t amend running task task!'


def test_stop_event_amend_message_reuse(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'stop', '--amend'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert f.read().endswith('stop message\n')


def test_stop_event_running_interactive(monkeypatch, tmpdir):
    monkeypatch.setattr('click.edit',
                        lambda s, **kwargs: 'interactive message')
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, '--interactive',
                                 'stop'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert f.read().endswith('interactive message\n')


def test_switch_event(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'switch', 'task2'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task2.csv').open() as f:
        assert len(f.read().splitlines()) == 3


def test_switch_event_with_file_message(tmpdir):
    test_dir = tmpdir.join('test').strpath
    msg_file = tmpdir.join('message').strpath
    with open(msg_file, 'w') as f:
        f.write('stopping message')
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'switch', '-F',
                                 msg_file, 'task2'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert f.read().endswith('stopping message\n')
    with tmpdir.join('test', 'task2.csv').open() as f:
        assert len(f.read().splitlines()) == 3


def test_switch_event_not_running(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'switch', 'task2'])
    assert isinstance(result.exception, TaskNotRunningError)
    assert result.exception.args[0] == 'No task running!'


def test_switch_start_date_overlaps(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'switch', '--time',
                                 '2011-05-04T09:15:00Z', 'task2'])
    assert isinstance(result.exception, TaskNotRunningError)
    assert result.exception.args[0] \
        == 'Can’t specify a start time before current task started!'


def test_switch_event_running_interactive(monkeypatch, tmpdir):
    monkeypatch.setattr('click.edit',
                        lambda s, **kwargs: 'interactive message')
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, '--interactive',
                                 'switch', 'task2'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert f.read().endswith('interactive message\n')
    with tmpdir.join('test', 'task2.csv').open() as f:
        assert len(f.read().splitlines()) == 3


def test_run_timed(capfd, tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'run', '-c',
                                 'echo "long running command"', 'task'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    assert capfd.readouterr()[0] == 'long running command\n'


def test_run_event_already_running(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'run', 'task2'])
    assert isinstance(result.exception, TaskRunningError)
    assert result.exception.args[0] == 'Task task is already started!'


def test_run_failed_command(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'run', '-c',
                                 '( exit 50 )', 'task'])
    assert isinstance(result.exception, OSError)
    assert result.exception.args == (50, 'Command failed')


def test_run_with_file_message(capfd, tmpdir):
    test_dir = tmpdir.join('test').strpath
    msg_file = tmpdir.join('message').strpath
    with open(msg_file, 'w') as f:
        f.write('stopping message')
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, 'run', '-F',
                                 msg_file, '-c',
                                 'echo "long running command"', 'task'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert f.read().endswith('stopping message\n')
    assert capfd.readouterr()[0] == 'long running command\n'


def test_run_interactive(monkeypatch, tmpdir):
    monkeypatch.setattr('click.edit',
                        lambda s, **kwargs: 'interactive message')
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', test_dir, '--interactive',
                                 'run', '-c', 'echo "long running command"',
                                 'task'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert f.read().endswith('interactive message\n')


def test_wrapper_no_config():
    runner = CliRunner()
    result = runner.invoke(cli, ['--config', 'tests/data/defaults.ini',
                                 'wrapper', 'calendar'])
    assert isinstance(result.exception, SystemExit)
    assert "No such wrapper 'calendar'" in result.output


def test_wrapper_run_command(capfd, tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    runner = CliRunner()
    result = runner.invoke(cli, ['--config', 'tests/data/wrappers.ini',
                                 '--directory', test_dir,
                                 'wrapper', 'calendar'])
    assert result.exit_code == 0
    assert 'Task task running for' in result.output
    with tmpdir.join('test', 'task.csv').open() as f:
        assert len(f.read().splitlines()) == 4
    assert 'May 2011' in capfd.readouterr()[0]


def test_report():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/test_not_running',
                                 'report'])
    assert result.exit_code == 0
    assert 'task    2:00:00' in result.output


def test_report_stats():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/test_not_running',
                                 'report', '--stats'])
    assert result.exit_code == 0
    assert 'Duration of events 2:15:00' in result.output


def test_report_event_running():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/test', 'report'])
    assert result.exit_code == 0
    assert 'Task “task” started 2011-05-04T09:30:00Z' \
        in result.output.splitlines()


@mark.parametrize('database, expected', [
    ('test', 'Task “task” started'),
    ('test_not_running', 'No task is running!'),
])
def test_running(database: str, expected: str):
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/' + database,
                                 'running'])
    assert result.exit_code == 0
    assert expected in result.output


@mark.parametrize('database, expected', [
    ('test', 'Task task is still running'),
    ('test_not_running', 'Last task task, ran for 1:00:00'),
])
def test_last(database: str, expected: str):
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/' + database,
                                 'last'])
    assert result.exit_code == 0
    assert expected in result.output


def test_ledger():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/test_not_running',
                                 'ledger'])
    assert result.exit_code == 0
    assert '2011-05-04 * 09:30-10:30' in result.output


def test_ledger_running():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/test', 'ledger'])
    assert result.exit_code == 0
    assert ';; Running event not included in output!' in result.output


def test_timeclock():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/test_not_running',
                                 'timeclock'])
    assert result.exit_code == 0
    assert 'i 2011-05-04 09:30:00 task' in result.output.splitlines()
    assert 'o 2011-05-04 10:30:00  ; stop message' in \
        result.output.splitlines()


def test_timeclock_running():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory', 'tests/data/test',
                                 'timeclock'])
    assert result.exit_code == 0
    assert ';; Running event not included in output!' in result.output


def test_main_wrapper(monkeypatch, capsys):
    monkeypatch.setattr('sys.argv', ['rdial', '--directory', 'tests/data/test',
                                     'running'])

    def exit_mock(n):
        if n == 0:
            return n
        else:
            raise ValueError(n)
    monkeypatch.setattr('sys.exit', exit_mock)
    result = main()
    assert result == 0
    assert 'Task “task” started' in capsys.readouterr()[0]


def test_main_wrapper_error(monkeypatch, capsys):
    monkeypatch.setattr('sys.argv', ['rdial', '--directory',
                                     'tests/data/test_not_running', 'stop'])

    def exit_mock(n):
        if n == 0:
            return n
        else:
            raise ValueError(n)
    monkeypatch.setattr('sys.exit', exit_mock)
    result = main()
    assert result == 2
    assert capsys.readouterr()[0] == 'No task running!\n'


def test_main_wrapper_informative_return_code(capsys, monkeypatch, tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    monkeypatch.setattr('sys.argv', ['rdial', '--directory', test_dir, 'run',
                                     '-c', '( exit 50 )', 'task'])
    result = main()
    assert result == 50
    assert 'Task task running for' in capsys.readouterr()[0]



#
"""utils - Utility functions for rdial."""
# Copyright © 2012-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import functools
import os
import subprocess
from datetime import date, datetime, timedelta
from typing import Callable, Dict, Optional, Tuple, Union

import click

from jnrbase import xdg_basedir
from jnrbase.iso_8601 import parse_datetime

try:
    import cduration
except ImportError:  # pragma: no cover
    cduration = None


class RdialError(ValueError):

    """Generic exception for rdial."""


#: Map duration string keys to timedelta args
_MAPPER = {'D': 'days', 'H': 'hours', 'M': 'minutes', 'S': 'seconds'} \
    # type : Dict[str, str]


def parse_datetime_user(__string: str) -> datetime:
    """Parse datetime string from user.

    We accept the normal ISO-8601 formats, but kick through to the formats
    supported by the system’s date command if parsing fails.

    Args:
        __string: Datetime string to parse

    Returns:
        Parsed datetime object

    """
    try:
        datetime_ = parse_datetime(__string)
    except ValueError:
        try:
            proc = subprocess.run(['date', '--utc', '--iso-8601=seconds',
                                   '-d', __string],
                                  stdout=subprocess.PIPE, check=True)
            output = proc.stdout.decode()
            datetime_ = parse_datetime(output.strip()[:19])
        except subprocess.CalledProcessError:
            datetime_ = None
    if not datetime_:
        raise ValueError(f'Unable to parse timestamp {__string!r}')
    return datetime_.replace(tzinfo=None)


def iso_week_to_date(__year: int, __week: int) -> Tuple[datetime.date,
                                                        datetime.date]:
    """Generate date range for a given ISO-8601 week.

    ISO-8601 defines a week as Monday to Sunday, with the first week of a year
    being the first week containing a Thursday.

    Args:
        __year: Year to process
        __week: Week number to process

    Returns:
        Date range objects for given week
    """
    bound = date(__year, 1, 4)
    iso_start = bound - timedelta(days=bound.isocalendar()[2] - 1)
    start = iso_start + timedelta(weeks=__week - 1)
    end = start + timedelta(weeks=1)
    return start, end


def read_config(user_config: Optional[str] = None,
                cli_options: Optional[Dict[str, Union[bool, str]]] = None
                ) -> configparser.ConfigParser:
    """Read configuration data.

    Args:
        user_config: User defined config file
        cli_options: Command line options

    Returns:
        Parsed configuration data

    """
    # Only base *must* exist
    conf = configparser.ConfigParser()
    # No, it *really* must
    with open(os.path.dirname(__file__) + '/config') as f:
        conf.read_file(f)
    conf['DEFAULT'] = {'xdg_data_location': xdg_basedir.user_data('rdial')}
    for f in xdg_basedir.get_configs('rdial'):
        conf.read(f)
    conf.read(os.path.abspath('.rdialrc'))
    if user_config:
        conf.read(user_config)

    if cli_options:
        conf.read_dict({
            'rdial': {k: v for k, v in cli_options.items() if v is not None}
        })

    return conf


def write_current(__fun: Callable) -> Callable:
    """Decorator to write :file:`.current` file on function exit.

    See also:
        :doc:`/taskbars`

    Args:
        __fun: Function to add hidden attribute to

    Returns:
        Wrapped function
    """
    @functools.wraps(__fun)
    def wrapper(*args, **kwargs):
        """Write value of ``task`` argument to ``.current`` on exit.

        Args:
            args: Positional arguments
            kwargs: Keyword arguments

        """
        globs = args[0]
        __fun(*args, **kwargs)
        with click.open_file(f'{globs.directory}/.current', 'w') as f:
            f.write(kwargs['task'])
    return wrapper


def remove_current(__fun: Callable) -> Callable:
    """Decorator to remove :file:`.current` file on function exit.

    See also:
        :doc:`/taskbars`

    Args:
        __fun: Function to add hidden attribute to

    Returns:
        Wrapped function
    """
    @functools.wraps(__fun)
    def wrapper(*args, **kwargs):
        """Remove ``.current`` file on exit.

        Args:
            args: Positional arguments
            kwargs: Keyword arguments

        """
        globs = args[0]
        __fun(*args, **kwargs)
        if os.path.isfile(f'{globs.directory}/.current'):
            os.unlink(f'{globs.directory}/.current')
    return wrapper


def newer(__fname: str, __reference: str) -> bool:
    """Check whether given file is newer than reference file.

    Args:
        __fname: File to check
        __reference: file to test against

    Returns:
        ``True`` if ``__fname`` is newer than ``__reference``

    """
    return os.stat(__fname).st_mtime > os.stat(__reference).st_mtime


def term_link(__target: str, name: Optional[str] = None) -> str:
    """Generate a terminal hyperlink

    See https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda.

    Args:
        __target: Hyperlink target
        name: Target name

    Returns:
        str: Formatted hyperlink for terminal output
    """
    if not name:
        name = os.path.basename(__target)
    return f'\033]8;;{__target}\007{name}\033]8;;\007'



#
"""conf - Sphinx configuration information."""
# Copyright © 2012-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from contextlib import suppress
from subprocess import CalledProcessError, PIPE, run

root_dir = os.path.dirname(os.path.dirname(__file__))
sys.path.insert(0, root_dir)

import rdial  # NOQA: E402

on_rtd = 'READTHEDOCS' in os.environ
if not on_rtd:
    import sphinx_rtd_theme

extensions = \
    [f'sphinx.ext.{ext}' for ext in ['autodoc', 'coverage', 'doctest',
                                     'intersphinx', 'napoleon', 'todo',
                                     'viewcode']] \
    + [f'sphinxcontrib.{ext}' for ext in []] \
    + ['sphinx_autodoc_typehints', ]  # type: List[str]


if not on_rtd:
    # Only activate spelling if it is installed.  It is not required in the
    # general case and we don’t have the granularity to describe this in a
    # clean way
    try:
        from sphinxcontrib import spelling  # NOQA: E401
    except ImportError:
        pass
    else:
        extensions.append('sphinxcontrib.spelling')

master_doc = 'index'
source_suffix = '.rst'

project = 'rdial'
copyright = '2011-2017  James Rowe'

release = rdial._version.dotted
version = release.rsplit('.', 1)[0]

html_experimental_html5_writer = True
modindex_common_prefix = ['rdial.', ]

# readthedocs.org handles this setup for their builds, but it is nice to see
# approximately correct builds on the local system too
if not on_rtd:
    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path(), ]  \
        # type: List[str]

pygments_style = 'sphinx'
with suppress(CalledProcessError):
    proc = run(['git', 'log', "--pretty=format:'%ad [%h]'", '--date=short',
                '-n1'],
               stdout=PIPE)
    html_last_updated_fmt = proc.stdout.decode()

man_pages = [
    ('rdial.1', 'rdial', 'rdial Documentation', ['James Rowe', ], 1)
]  # type: Tuple[str, str, str, List[str], int]

# Autodoc extension settings
autoclass_content = 'init'
autodoc_default_flags = ['members', ]  # type: List[str]

# intersphinx extension settings
intersphinx_mapping = {
    k: (v, os.getenv(f'SPHINX_{k.upper()}_OBJECTS'))
    for k, v in {
        'click': 'http://click.pocoo.org/6/',
        'jnrbase': 'http://jnrbase.readthedocs.io/en/latest/',
        'python': 'https://docs.python.org/3/',
}.items()}  # type: Dict[str, str]

# spelling extension settings
spelling_lang = 'en_GB'
spelling_word_list_filename = 'wordlist.txt'

# napoleon extension settings
napoleon_numpy_docstring = False

# todo extension settings
todo_include_todos = True



#
"""test_event - Test database handling"""
# Copyright © 2012-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.


from datetime import datetime, timedelta, timezone
from filecmp import dircmp
from glob import glob
from shutil import copytree
from typing import Optional, Union

from jnrbase.iso_8601 import parse_datetime, parse_delta
from pytest import mark, raises

from rdial import events as events_mod
from rdial.events import Event, Events, TaskRunningError


@mark.parametrize('task, start, delta, message', [
    ('test', None, None, None),
    ('test', '2013-02-26T19:45:14', None, None),
    ('test', datetime(2013, 2, 26, 19, 45, 14), None,
     None),
    ('test', '2013-02-26T19:45:14', 'PT8M19S', None),
    ('test', '2013-02-26T19:45:14', timedelta(minutes=8, seconds=19.770869),
     None),
    ('test', '2013-02-26T19:45:14', 'PT8M19S', 'stopped'),
])
def test_event_creation(task: str, start: Optional[Union[str, datetime]],
                        delta: Optional[Union[str, timedelta]], message: str):
    e = Event(task, start, delta, message)
    assert e.task == task
    if isinstance(start, datetime):
        assert e.start == start
    elif start:
        assert e.start == parse_datetime(start).replace(tzinfo=None)
    else:
        # Special case to ignore comparison against utcnow()
        pass
    if isinstance(delta, timedelta):
        assert e.delta == delta
    else:
        assert e.delta == parse_delta(delta)
    assert e.message == message


def test_event_creation_non_naive():
    with raises(ValueError, match='Must be a naive datetime'):
        Event('test', datetime(2013, 2, 26, 19, 45, 14, tzinfo=timezone.utc),
              None, None)


def test_event_equality():
    ev1 = Event('test', datetime(2013, 2, 26, 19, 45, 14),
                None, None)
    ev2 = Event('test', datetime(2013, 2, 26, 19, 45, 14),
                None, None)
    assert ev1 == ev2


@mark.parametrize('ev1, ev2', [
    (Event('test', datetime(2013, 2, 26, 19, 45, 14), ),
     Event('not_test', datetime(2013, 2, 26, 19, 45, 14), )),
    (Event('date', datetime(2013, 2, 26, 19, 45, 14), ),
     Event('date', datetime(2001, 1, 1, 0, 0, 0), )),
    (Event('message', datetime(2013, 2, 26, 19, 45, 14), None, 'test'),
     Event('message', datetime(2013, 2, 26, 19, 45, 14), None, 'breakage')),
])
def test_event_inequality(ev1: Event, ev2: Event):
    assert ev1 != ev2


@mark.parametrize('database, events', [
    ('test', 3),
    ('date_filtering', 3),
    ('test_not_running', 3),
])
def test_read_datebase(database: str, events: int):
    evs = Events.read('tests/data/' + database, write_cache=False)
    assert len(evs) == events


@mark.parametrize('database, events', [
    ('test', 3),
    ('date_filtering', 3),
    ('test_not_running', 3),
])
def test_read_datebase_wrapper(database: str, events: int):
    with Events.wrapping('tests/data/' + database, write_cache=False) as evs:
        assert len(evs) == events


def test_read_datebase_wrapper_write(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test', test_dir)
    with Events.wrapping(test_dir, write_cache=False) as evs:
        evs.stop()
    comp = dircmp('tests/data/test', test_dir, [])
    assert comp.diff_files == ['task.csv', ]
    assert comp.left_only == []
    assert comp.right_only == ['task.csv~', ]
    assert comp.funny_files == []
    assert comp.subdirs == {}


@mark.parametrize('database, result', [
    ('test', Event('task', '2011-05-04T09:30:00Z', '', 'finished')),
    ('', None),
])
def test_read_last(database: str, result: Optional[Event]):
    evs = Events.read('tests/data/' + database, write_cache=False)
    assert evs.last() == result


def test_fail_start_with_overlap():
    evs = Events.read('tests/data/test_not_running', write_cache=False)
    with raises(TaskRunningError, match='Start date overlaps'):
        evs.start('task', start=datetime(2011, 5, 4, 9, 33))


@mark.parametrize('n, task, start, delta', [
    (0, 'task', datetime(2011, 5, 4, 8),
     timedelta(hours=1)),
    (1, 'task2', datetime(2011, 5, 4, 9, 15),
     timedelta(minutes=15)),
    (2, 'task', datetime(2011, 5, 4, 9, 30), timedelta()),
])
def test_check_events(n: int, task: str, start: datetime, delta: timedelta):
    # FIXME: Clean-ish way to perform check, with the caveat that it parses the
    # database on each entry.  Need a better solution.
    events = Events.read('tests/data/test', write_cache=False)
    assert events[n].task == task
    assert events[n].start == start
    assert events[n].delta == delta


def test_write_database(tmpdir):
    in_dir = 'tests/data/test'
    events = Events.read(in_dir, write_cache=False)
    events._dirty = events.tasks()
    events.write(tmpdir.strpath)
    comp = dircmp(in_dir, tmpdir.strpath, [])
    assert comp.diff_files == []
    assert comp.left_only == []
    assert comp.right_only == []
    assert comp.funny_files == []
    assert comp.subdirs == {}


def test_write_database_event_backups(tmpdir):
    test_dir = tmpdir.join('test').strpath
    copytree('tests/data/test_not_running', test_dir)
    events = Events.read(test_dir, write_cache=False)
    events.start('task')
    events.write(test_dir)
    comp = dircmp('tests/data/test_not_running', test_dir, [])
    assert comp.diff_files == ['task.csv', ]
    assert comp.left_only == []
    assert comp.right_only == ['task.csv~', ]
    assert comp.funny_files == []
    assert comp.subdirs == {}


def test_write_database_no_change_noop(tmpdir):
    in_dir = 'tests/data/test'
    events = Events.read(in_dir, write_cache=False)
    events.write(tmpdir.strpath)
    assert glob(tmpdir.join('*').strpath) == []


def test_write_database_cache(monkeypatch, tmpdir):
    monkeypatch.setattr(events_mod.xdg_basedir, 'user_cache',
                        lambda s: tmpdir.join('cache').strpath)
    events = Events.read('tests/data/test')
    events._dirty = events.tasks()
    events.write(tmpdir.join('database').strpath)
    cache_files = glob(tmpdir.join('cache', '**', '*.pkl').strpath,
                       recursive=True)
    assert {f.split('/')[-1][:-4] for f in cache_files} == set(events.tasks())


def test_read_database_cache(monkeypatch, tmpdir):
    monkeypatch.setattr(events_mod.xdg_basedir, 'user_cache',
                        lambda s: tmpdir.join('cache').strpath)
    read = set()
    monkeypatch.setattr(
        events_mod.pickle, 'load',
        lambda f: read.add(f.name.split('/')[-1][:-4]))
    in_dir = 'tests/data/test'
    events = Events.read(in_dir)
    events = Events.read(in_dir)
    assert read == set(events.tasks())


def test_read_database_cache_broken(monkeypatch, tmpdir):
    monkeypatch.setattr(events_mod.xdg_basedir, 'user_cache',
                        lambda s: tmpdir.join('cache').strpath)
    in_dir = 'tests/data/test'
    events = Events.read(in_dir)
    cache_files = glob(tmpdir.join('cache', '**', '*.pkl').strpath,
                       recursive=True)
    with open(cache_files[0], 'w') as f:
        f.write('Broken data')
    events2 = Events.read(in_dir)
    assert events == events2


def test_store_messages_with_events():
    events = Events.read('tests/data/test', write_cache=False)
    assert events.last().message == 'finished'


def test_non_existing_database():
    assert Events() == Events.read('I_NEVER_EXIST', write_cache=False)



#
"""test_event_triggers - Test event trigger handling"""
# Copyright © 2012-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.


from pytest import raises

from rdial.events import (Events, TaskNotExistError, TaskNotRunningError,
                          TaskRunningError)


def test_start_event():
    events = Events.read('tests/data/test_not_running', write_cache=False)
    events.start('task2')
    assert events.running() == 'task2'


def test_fail_start_when_task_typo():
    events = Events.read('tests/data/test_not_running', write_cache=False)
    with raises(TaskNotExistError,
                match='Task non_existent does not exist!  Use “--new” to '
                      'create it'):
        events.start('non_existent')


def test_fail_start_when_running():
    events = Events.read('tests/data/test', write_cache=False)
    with raises(TaskRunningError, match='Running task task!'):
        events.start('task2')


def test_stop_event():
    events = Events.read('tests/data/test', write_cache=False)
    events.stop()
    assert not events.running()


def test_stop_event_with_message():
    events = Events.read('tests/data/test', write_cache=False)
    events.stop(message='test')
    last = events.last()
    assert last.message == 'test'


def test_fail_stop_when_not_running():
    events = Events.read('tests/data/test_not_running', write_cache=False)
    with raises(TaskNotRunningError, match='No task running!'):
        events.stop()


def test_fail_stop_single_when_not_running():
    events = Events.read('tests/data/test_not_running', write_cache=False)
    with raises(TaskNotRunningError, match='No task running!'):
        events.last().stop()



#
"""cmdline - Command line functionality for rdial."""
# Copyright © 2012-2018  James Rowe <jnrowe@gmail.com>
#                        Nathan McGregor <nathan.mcgregor@astrium.eads.net>
#                        Rach Holmes <rach.c.holmes@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import operator
import os
import shlex
import subprocess
from typing import Callable, List, Optional

import click
import tabulate

from jnrbase import colourise, i18n, iso_8601
from jnrbase.attrdict import AttrDict

from . import _version, utils
from .events import Event, Events, TaskNotRunningError, TaskRunningError

_, N_ = i18n.setup(_version)


class HiddenGroup(click.Group):

    """Support for 'hidden' commands.

    Any :obj:`click.Command` with the hidden attribute set will not be visible
    in help output.  This is mainly to be used for diagnostic commands, and
    shouldn’t be abused!

    """

    def list_commands(self, __ctx: click.Context) -> List[str]:
        """List visible commands.

        Args:
            __ctx: Current command context
        Returns:
            Visible command names
        """
        return sorted(k for k, v in self.commands.items()
                      if not hasattr(v, 'hidden'))


class TaskNameParamType(click.ParamType):

    """Task name parameter handler."""

    name = 'taskname'

    def convert(self, __value: str, __param: Optional[click.Argument],
                __ctx: Optional[click.Context]) -> str:
        """Check given task name is valid.

        Args:
            __value: Value given to flag
            __param: Parameter being processed
            __ctx: Current command context

        Returns:
            Valid task name

        """
        if not __value:
            raise click.BadParameter(_('No task name given'))
        if __value.startswith('-'):
            raise click.BadParameter(_('Task names with leading dashes are '
                                       'non-portable'))
        if __value.startswith('.') or '/' in __value or '\000' in __value:
            raise click.BadParameter(
                _(f'{__value!r} is not a valid task name'))
        # Should be based on platform’s PATH_MAX, but it isn’t exposed in a
        # clean way to Python
        if len(__value) > 255:
            raise click.BadParameter(
                _(f'{__value!r} is too long to be a valid task name(max 255 '
                  'characters)'))
        return __value


class StartTimeParamType(click.ParamType):

    """Start time parameter handler."""

    name = 'time'

    def convert(self, __value: str, __param: Optional[click.Argument],
                __ctx: Optional[click.Context]) -> datetime.datetime:
        """Check given start time is valid.

        Args:
            __value: Value given to flag
            __param: Parameter being processed
            __ctx: Current command context

        Returns:
            Valid start time

        """
        try:
            __value = utils.parse_datetime_user(__value)
        except ValueError:
            raise click.BadParameter(
                _(f'{__value!r} is not a valid ISO-8601 time string'))
        return __value


def task_from_dir(__ctx: click.Context, __param: click.Option,
                  __value: bool) -> None:
    """Override task name default using name of current directory.

    Args:
        __ctx: Current command context
        __param: Parameter being processed
        __value: True if flag given

    """
    if not __value or __ctx.resilient_parsing:
        return
    __param = [p for p in __ctx.command.params if p.name == 'task'][0]
    __param.default = os.path.basename(os.path.abspath(os.curdir))


def get_stop_message(__current: Event, __edit: bool = False) -> str:
    """Interactively fetch stop message.

    Args:
        __current: Current task
        __edit: Whether to edit existing message

    Returns:
        Message to use

    """
    marker = _('# Text below here ignored\n')
    task_message = _(f'# Task “{__current.task}” started '
                     f'{iso_8601.format_datetime(__current.start)}Z')
    template = f'{__current.message}\n{marker}{task_message}'
    message = click.edit(template, require_save=not __edit)
    if message is None:
        message = ''
    else:
        message = message.split(marker, 1)[0].rstrip('\n')
    return message


def task_option(__fun: Callable) -> Callable:
    """Add task selection options.

    Note:
        This is only here to reduce duplication in command setup.

    Args:
        __fun: Function to add options to

    Returns:
        Function with additional options

    """
    __fun = click.option('-x', '--from-dir', is_flag=True, expose_value=False,
                         is_eager=True, callback=task_from_dir,
                         help=_('Use directory name as task name.'))(__fun)
    __fun = click.argument('task', default='default', envvar='RDIAL_TASK',
                           required=False, type=TaskNameParamType())(__fun)
    return __fun


def duration_option(__fun: Callable) -> Callable:
    """Add duration selection option.

    Note:
        This is only here to reduce duplication in command setup.

    Args:
        __fun: Function to add options to

    Returns:
        Function with additional options

    """
    __fun = click.option(
        '-d', '--duration', default='all',
        type=click.Choice(['day', 'week', 'month', 'year', 'all']),
        help=_('Filter events for specified time period.'))(__fun)
    return __fun


def message_option(__fun: Callable) -> Callable:
    """Add message setting options.

    Note:
        This is only here to reduce duplication in command setup.

    Args:
        __fun: Function to add options to

    Returns:
        Function with additional options

    """
    __fun = click.option('-m', '--message', help=_('Closing message.'))(__fun)
    __fun = click.option('-F', '--file', 'fname', type=click.File(),
                         help=_('Read closing message from file.'))(__fun)
    return __fun


def hidden(__fun: click.Command) -> click.Command:
    """Add a hidden attribute to a Command object.

    :see:`HiddenGroup`.

    Args:
        __fun: Function to add hidden attribute to

    Returns:
        Function with hidden attribute set

    """
    __fun.hidden = True
    return __fun


# pylint: disable=too-many-arguments

@click.group(cls=HiddenGroup,
             help=_('Simple time tracking for simple people.'),
             epilog=_('Please report bugs at '
                      'https://github.com/JNRowe/rdial/issues'),
             context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(_version.dotted)
@click.option('-d', '--directory', metavar='DIR',
              type=click.Path(file_okay=False),
              help=_('Directory to read/write to.'))
@click.option('--backup/--no-backup', default=None,
              help=_('Do not write data file backups.'))
@click.option('--cache/--no-cache', default=None,
              help=_('Do not write cache files.'))
@click.option('--config', type=click.Path(exists=True, dir_okay=False,
                                          resolve_path=True, allow_dash=True),
              help=_('File to read configuration data from.'))
@click.option('-i', '--interactive/--no-interactive', default=None,
              help=_('Support interactive message editing.'))
@click.option('--colour/--no-colour', envvar='RDIAL_COLOUR', default=None,
              help=_('Output colourised informational text.'))
@click.pass_context
def cli(ctx: click.Context, directory: str, backup: bool, cache: bool,
        config: str, interactive: bool, colour: bool):
    """Main command entry point.

    Args:
        ctx: Current command context
        directory: Location to store event data
        backup: Whether to create backup files
        cache: Whether to create cache files
        config: Location of config file
        interactive: Whether to support interactive message editing
        colour: Whether to colourise output

    """
    cli_options = {
        'backup': backup,
        'cache': cache,
        'directory': directory,
        'interactive': interactive,
    }

    cfg = utils.read_config(config, cli_options)

    base = cfg['rdial']

    if colour is None:
        if 'color' in base:
            base['colour'] = base['color']
        colour = base.getboolean('colour')
    colourise.COLOUR = colour

    ctx.default_map = {}
    for name in ctx.command.commands:
        if name in cfg.sections():
            defs = {}
            for k in cfg[name]:
                try:
                    defs[k] = cfg[name].getboolean(k)
                except ValueError:
                    defs[k] = cfg[name][k]
            ctx.default_map[name] = defs

    ctx.obj = AttrDict(
        backup=base.getboolean('backup'),
        cache=base.getboolean('cache'),
        colour=colour,
        config=cfg,
        directory=base['directory'],
        interactive=base.getboolean('interactive'),
    )


def filter_events(__globs: AttrDict, __task: Optional[str] = None,
                  __duration: str = 'all') -> Events:
    """Filter events for report processing.

    Args:
        __globs: Global options object
        __task: Task name to filter on
        __duration: Time window to filter on

    Returns:
        Events: Events matching specified criteria

    """
    events = Events.read(__globs.directory, write_cache=__globs.cache)
    if __task:
        events = events.for_task(__task)
    if not __duration == 'all':  # pragma: no cover
        if __duration == 'week':
            today = datetime.date.today()
            events = events.for_week(*today.isocalendar()[:2])
        else:
            year, month, day = datetime.date.today().timetuple()[:3]
            if __duration == 'month':
                day = None
            elif __duration == 'year':
                month = None
                day = None
            events = events.for_date(year, month, day)
    return events


@hidden
@cli.command('bug-data', help=_('Produce data for rdial bug reports.'))
def bug_data():
    """Produce data for rdial bug reports."""
    import sys

    from pkg_resources import (DistributionNotFound, get_distribution)

    click.echo(f'* OS: {sys.platform}')
    click.echo(f'* `rdial` version: {_version.dotted}')
    click.echo('* `python` version: {}'.format(sys.version.replace('\n', '|')))
    click.echo()

    for m in ['click', 'cduration', 'jnrbase', 'tabulate']:
        try:
            pkg = get_distribution(m)
        except DistributionNotFound:
            continue
        link = utils.term_link(
            f'https://pypi.org/project/{pkg.project_name}/',
            f'`{pkg.project_name}`')
        click.echo(f'* {link}: {pkg.version}')


@cli.command(help=_('Check storage consistency.'))
@click.pass_obj
@click.pass_context
def fsck(ctx: click.Context, globs: AttrDict):
    """Check storage consistency.

    Args:
        ctx: Current command context
        globs: Global options object

    """
    warnings = 0
    events = Events.read(globs.directory, write_cache=globs.cache)
    now = datetime.datetime.utcnow()
    lines = []
    # Note: progress is *four* times slower on my data and system
    with click.progressbar(events, label=_('Checking'),
                           fill_char=click.style(u'█', 'green'),
                           empty_char=click.style(u'·', 'yellow')) as pbar:
        last_event = next(pbar)
        for event in pbar:
            if not last_event.start + last_event.delta <= event.start:
                warnings += 1
                lines.append(colourise.fail(_('Overlap:')))
                lines.append(colourise.warn(f'   {last_event!r}'))
                lines.append(colourise.info(f'   {event!r}'))
            if event.start > now:
                warnings += 1
                lines.append(colourise.fail(_('Future start:')))
                lines.append(colourise.warn(f'   {event!r}'))
            elif event.start + event.delta > now:
                warnings += 1
                lines.append(colourise.fail(_('Future end:')))
                lines.append(colourise.warn(f'   {event!r}'))
            last_event = event
    if lines:
        click.echo_via_pager('\n'.join(lines))
        if warnings:
            # Will be success when 𝐱 % 256 == 0, so cap at 255.  That said
            # you’ve got bigger problems if you’re hitting this ;)
            ctx.exit(min(warnings, 255))


@cli.command(help=_('Start task.'))
@task_option
@click.option('-c', '--continue', 'continue_', is_flag=True,
              help=_('Restart previous task.'))
@click.option('-n', '--new', is_flag=True, help=_('Start a new task.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@click.pass_obj
@utils.write_current
def start(globs: AttrDict, task: str, continue_: bool, new: bool,
          time: datetime):
    """Start task.

    Args:
        globs: Global options object
        task: Task name to operate on
        continue_: Pull task name from last running task
        new: Create a new task
        time: Task start time

    """
    with Events.wrapping(globs.directory, globs.backup, globs.cache) as events:
        if continue_:
            task = events.last().task
        events.start(task, new, time)


@cli.command(help=_('Stop task.'))
@message_option
@click.option('--amend', is_flag=True, help=_('Amend previous stop entry.'))
@click.pass_obj
@utils.remove_current
def stop(globs: AttrDict, message: str, fname: str, amend: bool):
    """Stop task.

    Args:
        globs: Global options object
        message: Message to assign to event
        fname: Filename to read message from
        amend: Amend a previously stopped event

    """
    if fname:
        message = fname.read()
    with Events.wrapping(globs.directory, globs.backup, globs.cache) as events:
        last_event = events.last()
        if last_event.running():
            if amend:
                raise TaskRunningError(
                    _(f'Can’t amend running task {last_event.task}!'))
        else:
            if not amend:
                raise TaskNotRunningError(_('No task running!'))
        if amend and not message:
            message = last_event.message
        if globs.interactive and not message:
            message = get_stop_message(last_event, amend)
        events.stop(message, force=amend)
    event = events.last()
    click.echo(_('Task {} running for {}').format(
        event.task,
        str(event.delta).split('.')[0]))


@cli.command(help=_('Switch to another task.'))
@task_option
@click.option('-n', '--new', is_flag=True, help=_('Start a new task.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@message_option
@click.pass_obj
@utils.write_current
def switch(globs: AttrDict, task: str, new: bool, time: datetime,
           message: str, fname: str):
    """Complete last task and start new one.

    Args:
        globs: Global options object
        task: Task name to operate on
        new: Create a new task
        time: Task start time
        message: Message to assign to event
        fname: Filename to read message from

    """
    if fname:
        message = fname.read()
    with Events.wrapping(globs.directory, globs.backup, globs.cache) as events:
        event = events.last()
        if time and time < event.start:
            raise TaskNotRunningError(_('Can’t specify a start time before '
                                        'current task started!'))
        if not event.running():
            raise TaskNotRunningError(_('No task running!'))
        if new or task in events.tasks():
            if globs.interactive and not message:
                message = get_stop_message(event)
            # This is dirty, but we kick on to Events.start() to save
            # duplication of error handling for task names
            events.stop(message)
        events.last().delta = time - event.start
        events.start(task, new, time)
    click.echo(_('Task {} running for {}').format(
        event.task,
        str(event.delta).split('.')[0]))


@cli.command(help=_('Run command with timer.'))
@task_option
@click.option('-n', '--new', is_flag=True, help=_('Start a new task.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@message_option
@click.option('-c', '--command', help=_('Command to run.'))
@click.pass_obj
def run(globs: AttrDict, task: str, new: bool, time: datetime, message: str,
        fname: str, command: str):
    """Run timed command.

    Args:
        globs: Global options object
        task: Task name to operate on
        new: Create a new task
        time: Task start time
        message: Message to assign to event
        fname: Filename to read message from
        command: Command to run

    """
    with Events.wrapping(globs.directory, globs.backup, globs.cache) as events:
        if events.running():
            raise TaskRunningError(
                _(f'Task {events.last().task} is already started!'))

        proc = subprocess.run(command, shell=True)

        events.start(task, new, time)
        with click.open_file(f'{globs.directory}/.current', 'w') as f:
            f.write(task)

        if fname:
            message = fname.read()
        if globs.interactive and not message:
            message = get_stop_message(events.last())
        events.stop(message)
    event = events.last()
    click.echo(_('Task {} running for {}').format(
        event.task,
        str(event.delta).split('.')[0]))
    os.unlink(f'{globs.directory}/.current')
    if proc.returncode != 0:
        raise OSError(proc.returncode, _('Command failed'))


@cli.command(help=_('Run predefined command with timer.'))
@click.option('-t', '--time', default='', help=_('Set start time.'),
              type=StartTimeParamType())
@message_option
@click.argument('wrapper', default='default')
@click.pass_obj
@click.pass_context
def wrapper(ctx: click.Context, globs: AttrDict, time: datetime, message: str,
            fname: str, wrapper: str):
    """Run predefined timed command.

    Args:
        ctx: Click context object
        globs: Global options object
        time: Task start time
        message: Message to assign to event
        fname: Filename to read message from
        wrapper: Run wrapper to execute

    """
    try:
        command = globs.config['run wrappers'][wrapper]
    except KeyError:
        raise click.BadParameter(_(f'No such wrapper {wrapper!r}'))
    parser = ctx.parent.command.commands['run'].make_parser(ctx)
    args = {'time': time, 'message': message, 'fname': fname, 'new': False}
    args.update(parser.parse_args(shlex.split(command))[0])
    ctx.invoke(run, **args)  # pylint: disable=star-args


@cli.command(help=_('Report time tracking data.'))
@task_option
@click.option('--stats', is_flag=True,
              help=_('Display database statistics.'))
@duration_option
@click.option('-s', '--sort', default='task', envvar='RDIAL_SORT',
              type=click.Choice(['task', 'time']), help=_('Field to sort by.'))
@click.option('-r', '--reverse/--no-reverse', default=False,
              envvar='RDIAL_REVERSE',
              help=_('Reverse sort order.'))
@click.option('--style', default='simple',
              type=click.Choice(tabulate._table_formats.keys()),
              help=_('Table output style.'))
@click.pass_obj
def report(globs: AttrDict, task: str, stats: bool, duration: str, sort: str,
           reverse: bool, style: str):
    """Report time tracking data.

    Args:
        globs: Global options object
        task: Task name to operate on
        stats: Display short overview of data
        duration: Time window to filter on
        sort: Key to sort events on
        reverse: Reverse sort order
        style: Table formatting style

    """
    if task == 'default':
        # Lazy way to remove duplicate argument definitions
        task = None
    events = filter_events(globs, task, duration)
    if stats:
        click.echo(N_('{} event in query', '{} events in query',
                      len(events)).format(len(events)))
        click.echo(_(f'Duration of events {events.sum()}'))
        if events:
            click.echo(_('First entry started at {events[0].start}'))
            click.echo(_('Last entry started at {events[-1].start}'))
        dates = {e.start.date() for e in events}
        click.echo(_('Events exist on {len(dates)} dates'))
    else:
        data = sorted(([t, str(events.for_task(t).sum())]
                       for t in events.tasks()),
                      key=operator.itemgetter(['task', 'time'].index(sort)),
                      reverse=reverse)
        click.echo_via_pager(tabulate.tabulate(data, ['task', 'time'],
                                               tablefmt=style))
    if events.running():
        current = events.last()
        click.echo(_(f'Task “{current.task}” started '
                     f'{iso_8601.format_datetime(current.start)}Z'))


@cli.command(help=_('Display running task, if any.'))
@click.pass_obj
def running(globs: AttrDict):
    """Display running task, if any.

    Args:
        globs: Global options object

    """
    events = Events.read(globs.directory, write_cache=globs.cache)
    if events.running():
        current = events.last()
        now = datetime.datetime.utcnow()
        click.echo(_('Task “{}” started {}').format(
            current.task, str(now - current.start).split('.')[0]))
    else:
        colourise.pwarn(_('No task is running!'))


@cli.command(help=_('Display last event, if any.'))
@click.pass_obj
def last(globs: AttrDict):
    """Display last event, if any.

    Args:
        globs: Global options object

    """
    events = Events.read(globs.directory, write_cache=globs.cache)
    event = events.last()
    if not events.running():
        click.echo(_(f'Last task {event.task}, ran for {event.delta}'))
        if event.message:
            click.echo(event.message)
    else:
        colourise.pwarn(_(f'Task {event.task} is still running'))


@cli.command(help=_('Generate ledger compatible data file.'))
@task_option
@duration_option
@click.option('-r', '--rate', type=float, envvar='RDIAL_RATE',
              help=_('Hourly rate for task output.'))
@click.pass_obj
def ledger(globs: AttrDict, task: str, duration: str, rate: str):
    """Generate ledger compatible data file.

    Args:
        globs: Global options object
        task: Task name to operate on
        duration: Time window to filter on
        rate: Rate to assign hours in report

    """
    if task == 'default':
        # Lazy way to remove duplicate argument definitions
        task = None
    events = filter_events(globs, task, duration)
    lines = []
    if events.running():
        lines.append(_(';; Running event not included in output!'))
    for event in events:
        if not event.delta:
            continue
        end = event.start + event.delta
        hours = event.delta.total_seconds() / 3600
        lines.append(f'{event.start:%F * %H:%M}-{end:%H:%M}')
        lines.append('    (task:{})  {:.2f}h{}{}'.format(
            event.task, hours, ' @ {}'.format(rate) if rate else '',
            '  ; {}'.format(event.message) if event.message else ''))
    if events.running():
        lines.append(_(';; Running event not included in output!'))
    click.echo_via_pager('\n'.join(lines))


@cli.command(help=_('Generate ledger compatible timeclock file.'))
@task_option
@duration_option
@click.pass_obj
def timeclock(globs: AttrDict, task: str, duration: str):
    """Generate ledger compatible timeclock data file.

    Args:
        globs: Global options object
        task: Task name to operate on
        duration: Time window to filter on
    """
    if task == 'default':
        # Lazy way to remove duplicate argument definitions
        task = None
    events = filter_events(globs, task, duration)
    lines = []
    if events.running():
        lines.append(_(';; Running event not included in output!'))
    for event in events:
        if not event.delta:
            continue
        lines.append(f'i {event.start:%F %T} {event.task}')
        lines.append(f'o {event.start + event.delta:%F %T}'
                     f'{"  ; " + event.message if event.message else ""}\n')
    if events.running():
        lines.append(_(';; Running event not included in output!'))
    click.echo_via_pager('\n'.join(lines))


# pylint: enable=too-many-arguments


def main() -> int:
    """Command entry point to handle errors.

    Returns:
        Final exit code

    """
    try:
        # pylint: disable=no-value-for-parameter
        cli(auto_envvar_prefix='RDIAL')
        return 0
    except (ValueError, utils.RdialError) as error:
        colourise.pfail(str(error))
        return 2
    except OSError as error:
        return error.errno



#
"""test_fsck - Test database fsck support"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime

from click.testing import CliRunner
from hiro import Timeline

from rdial.cmdline import cli


def test_fsck_overlap():
    runner = CliRunner()
    result = runner.invoke(cli, ['--directory=tests/data/test_fsck',
                                 '--no-cache', 'fsck'])
    assert result.exit_code == 1
    assert 'Overlap' in result.output
    assert "'2011-05-04T09:15:00Z', 'PT35M'" in result.output


def test_fsck_future_start():
    with Timeline(start=datetime(2016, 12, 13, 23, 0)):
        runner = CliRunner()
        result = runner.invoke(cli, ['--directory=tests/data/test_fsck_future',
                                     '--no-cache', 'fsck'])
    assert result.exit_code == 1
    assert 'Future start' in result.output


def test_fsck_future_end():
    with Timeline(start=datetime(2016, 12, 13, 23, 7)):
        runner = CliRunner()
        result = runner.invoke(cli, ['--directory=tests/data/test_fsck_future',
                                     '--no-cache', 'fsck'])
    assert result.exit_code == 1
    assert 'Future end' in result.output



#
"""test_event_query - Test event query handling"""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of rdial.
#
# rdial is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# rdial is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# rdial.  If not, see <http://www.gnu.org/licenses/>.

from datetime import timedelta

from rdial.events import Events


def test_list_tasks():
    events = Events.read('tests/data/test', write_cache=False)
    assert events.tasks() == ['task', 'task2']


def test_current_running_event():
    events = Events.read('tests/data/test', write_cache=False)
    assert events.running() == 'task'


def test_no_currently_running_event():
    events = Events.read('tests/data/test_not_running', write_cache=False)
    assert not events.running()


def test_sum_durations_in_database():
    events = Events.read('tests/data/test_not_running', write_cache=False)
    assert events.sum() == timedelta(hours=2, minutes=15)
